from pathlib import Path
import pandas as pd
import numpy as np
from microdf import MicroDataFrame
from openfisca_uk_data.utils import dataset

NON_MONTHLY_VARIABLES = (
    "person_id",
    "benunit_id",
    "household_id",
    "person_weight",
    "lhw",
    "dag",
)


@dataset
class UKMODInput:
    name = "ukmod_input"

    def generate(tabfile: str, year: int):
        tabfile = Path(tabfile)
        # Read the input dataset and construct a weighted DataFrame
        df = pd.read_csv(tabfile, delimiter="\t")
        # Add IDs to match against OpenFisca-UK
        df["person_id"] = (
            df.idorighh * 1e2 + df.idorigbenunit * 1e1 + df.idorigperson
        ).astype(int)
        df["benunit_id"] = (df.idorighh * 1e2 + df.idorigbenunit * 1e1).astype(
            int
        )
        df["household_id"] = (df.idorighh * 1e2).astype(int)
        df["person_weight"] = df.dwt
        df.set_index("person_id", inplace=True)
        for variable in df.columns:
            if variable not in NON_MONTHLY_VARIABLES:
                df[variable] *= 12
        with pd.HDFStore(UKMODInput.file(year)) as f:
            f["person"] = pd.DataFrame(df)



from pathlib import Path
import shutil
from typing import Callable, Type
from openfisca_core.model_api import *
from pathlib import Path
import shutil
from typing import List
import pandas as pd
import re
import os
from functools import wraps
import h5py
import requests
from tqdm import tqdm
import numpy as np

UK = "openfisca_uk"


class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


def dataset(cls):
    def generate():
        raise NotImplementedError("No dataset generation function specified")

    if not hasattr(cls, "model"):
        cls.model = None
        cls.data_dir = data_folder(DATA_DIR / "external")
    else:
        cls.data_dir = data_folder(DATA_DIR / cls.model)

    def years(cl):
        pattern = re.compile(f"\n{cl.name}_([0-9]+).h5")
        matches = list(
            map(
                int,
                pattern.findall(
                    "\n"
                    + "\n".join(
                        map(lambda path: path.name, cl.data_dir.iterdir())
                    )
                ),
            )
        )
        return matches

    cls.years = classproperty(years)

    def last_year(self):
        return sorted(cls.years)[-1]

    cls.last_year = property(last_year)

    def filename(year):
        return f"{cls.name}_{year}.h5"

    cls.filename = staticmethod(filename)

    def load(year, key: str = None) -> pd.DataFrame:
        file = cls.file(year)
        if cls.model:
            if key is None:
                try:
                    return h5py.File(file, mode="a")
                except OSError:
                    return h5py.File(file, mode="r")
            else:
                with h5py.File(file, mode="r") as f:
                    values = np.array(f[key])
                return values
        else:
            if key is None:
                return pd.HDFStore(file)
            else:
                with pd.HDFStore(file) as f:
                    values = f[key]
                return values

    def remove(year=None):
        if year is None:
            filenames = map(cls.filename(year), cls.years)
        else:
            filenames = (cls.filename(year),)
        for filename in filenames:
            filepath = cls.data_dir / filename
            if filepath.exists():
                os.remove(filepath)

    cls.remove = staticmethod(remove)

    def remove_first_then(generate_func):
        def new_generate_func(year, *args):
            cls.remove(year)
            return generate_func(year, *args)

        return new_generate_func

    if hasattr(cls, "generate"):
        cls.generate = staticmethod(remove_first_then(cls.generate))
    else:
        cls.generate = staticmethod(generate)

    cls.load = staticmethod(load)

    if not hasattr(cls, "input_reform_from_year"):
        cls.input_reform_from_year = lambda year: ()

    cls.file = staticmethod(lambda year: cls.data_dir / cls.filename(year))

    def save(data_file: str, year: int = 2018):
        if "https://" in data_file:
            response = requests.get(data_file, stream=True)
            total_size_in_bytes = int(
                response.headers.get("content-length", 0)
            )
            block_size = 1024  # 1 Kibibyte
            progress_bar = tqdm(
                total=total_size_in_bytes, unit="iB", unit_scale=True
            )
            with open(cls.file(year), "wb") as file:
                for data in response.iter_content(block_size):
                    progress_bar.update(len(data))
                    file.write(data)
            progress_bar.close()
        else:
            shutil.copyfile(data_file, cls.file(year))

    if not hasattr(cls, "save"):
        cls.save = staticmethod(save)

    if not hasattr(cls, "upload"):

        def upload(year):
            from google.cloud import storage

            client = storage.Client()
            bucket = client.get_bucket("policyengine-uk-data")
            blob = bucket.blob(cls.file(year).name)
            with open(cls.file(year), "rb") as f:
                blob.upload_from_file(f)

        cls.upload = staticmethod(upload)

    if not hasattr(cls, "download"):

        def download(year):
            from google.cloud import storage

            client = storage.Client()
            bucket = client.get_bucket("policyengine-uk-data")
            blob = bucket.blob(cls.file(year).name)
            with open(cls.file(year), "wb") as f:
                blob.download_to_file(f)

        cls.download = staticmethod(download)

    return cls


def data_folder(path: str, erase=False) -> Path:
    folder = Path(path)
    folder.mkdir(exist_ok=True, parents=True)
    if erase:
        shutil.rmtree(folder)
        folder.mkdir()
    return folder


def safe_rmdir(path: str):
    if Path(path).exists():
        shutil.rmtree(path)


PACKAGE_DIR = Path(__file__).parent
DATA_DIR = PACKAGE_DIR / "microdata"



from openfisca_uk_data.utils import dataset
from openfisca_uk_data.datasets.frs.raw_frs import RawFRS
from openfisca_uk_data.datasets.frs.ukmod.ukmod_input import UKMODInput
from openfisca_uk_data.datasets.frs.ukmod.ukmod_output import UKMODOutput
from openfisca_uk_data.datasets.frs.frs import FRS
from openfisca_uk_data.datasets.frs.frs_spi_adjusted import FRS_SPI_Adjusted
from openfisca_uk_data.datasets.frs.frs_was_imputation import (
    FRS_WAS_Imputation,
)
from openfisca_uk_data.datasets.frs.synth_frs import SynthFRS



from openfisca_uk_data.utils import *
from openfisca_uk_data.datasets.frs.frs import FRS
import shutil
import pandas as pd
import numpy as np
import h5py
import requests
from tqdm import tqdm

DEFAULT_SYNTH_URL = "https://github.com/UBICenter/openfisca-uk-data/releases/download/synth-frs/synth_frs_2018.h5"


@dataset
class SynthFRS:
    name = "synth_frs"
    model = UK
    input_reform_from_year = FRS.input_reform_from_year

    def generate(year):
        from openfisca_uk import CountryTaxBenefitSystem

        ID_COLS = (
            "P_person_id",
            "P_benunit_id",
            "P_household_id",
            "B_benunit_id",
            "B_household_id",
            "H_household_id",
        )

        def anonymise(df: pd.DataFrame) -> pd.DataFrame:
            result = df.copy()
            for col in result.columns:
                if col not in ID_COLS:
                    # don't change identity columns, this breaks structures
                    if result[col].unique().size < 16:
                        # shuffle categorical columns
                        result[col] = result[col].sample(frac=1).values
                    else:
                        # shuffle + add noise to numeric columns
                        # noise = between -3% and +3% added to each row
                        noise = np.random.rand() * 3e-2 + 1.0
                        result[col] = result[col].sample(frac=1).values * noise
            return result

        year = 2018
        system = CountryTaxBenefitSystem()
        data = FRS.load(year)
        entities = ("person", "benunit", "household")
        entity_dfs = {key: {} for key in entities}
        for entity in entities:
            for variable in data.keys():
                if system.variables[variable].entity.key == entity:
                    entity_dfs[entity][variable] = data[variable][str(year)]
        person, benunit, household = map(
            lambda x: anonymise(pd.DataFrame(x)), entity_dfs.values()
        )

        year = int(year)

        with h5py.File(SynthFRS.file(year), mode="w") as f:
            for variable in person.columns:
                f[f"{variable}/{year}"] = person[variable].values
            for variable in benunit.columns:
                f[f"{variable}/{year}"] = benunit[variable].values
            for variable in household.columns:
                f[f"{variable}/{year}"] = household[variable].values

    def save(data_file: str = DEFAULT_SYNTH_URL, year: int = 2018):
        if "https://" in data_file:
            response = requests.get(data_file, stream=True)
            total_size_in_bytes = int(
                response.headers.get("content-length", 0)
            )
            block_size = 1024  # 1 Kibibyte
            progress_bar = tqdm(
                total=total_size_in_bytes, unit="iB", unit_scale=True
            )
            with open(SynthFRS.file(year), "wb") as file:
                for data in response.iter_content(block_size):
                    progress_bar.update(len(data))
                    file.write(data)
            progress_bar.close()
        else:
            shutil.copyfile(data_file, SynthFRS.file(year))



from openfisca_uk_data.utils import dataset, UK
from openfisca_uk_data.datasets.frs.frs import FRS
import pandas as pd
import microdf as mdf
import sklearn
import synthimpute as si
import h5py
import numpy as np


@dataset
class FRS_WAS_Imputation:
    name = "frs_was_imp"
    model = UK

    def generate(was_file: str, year: int):
        was_df = process_was(was_file)
        pred_land = impute_land(was_df, year)
        frs_was = h5py.File(FRS_WAS_Imputation.file(year), mode="w")
        frs = FRS.load(year)
        for variable in tuple(frs.keys()):
            frs_was[variable] = np.array(frs[variable][...])
        frs_was["land_value"] = pred_land
        frs_was.close()
        frs.close()


def impute_land(was_df: pd.DataFrame, year: int) -> pd.Series:
    """Impute land by fitting a random forest model.

    Args:
            was_df (pd.DataFrame): The WAS dataset.
            year (int): The year of simulation.

    Returns:
            pd.Series: The predicted land values.
    """
    from openfisca_uk import Microsimulation

    sim = Microsimulation(dataset=FRS, year=year)

    was = pd.read_csv("~/was.csv")

    TRAIN_COLS = [
        "gross_income",
        "num_adults",
        "num_children",
        "pension_income",
        "employment_income",
        "self_employment_income",
        "investment_income",
        "num_bedrooms",
        "council_tax",
        "is_renting",
        # "region"  # TODO: add region mapping from WAS int.
    ]

    IMPUTE_COLS = [
        "est_land",  # Estimated land value based on property and corporate wealth.
    ]

    # FRS has investment income split between dividend and savings interest.
    frs_cols = [i for i in TRAIN_COLS if i != "investment_income"]
    frs_cols += [
        "dividend_income",
        "savings_interest_income",
        "people",
        "net_income",
        "household_weight",
    ]

    frs = sim.df(frs_cols, map_to="household", period=year)
    frs["investment_income"] = (
        frs.savings_interest_income + frs.dividend_income
    )

    return si.rf_impute(
        x_train=was[TRAIN_COLS],
        y_train=was[IMPUTE_COLS],
        x_new=frs[TRAIN_COLS],
        sample_weight_train=was.weight,
        new_weight=frs.household_weight,
        target=mdf.weighted_sum(was, "est_land", "weight"),
    )


def process_was(was_file: str) -> pd.DataFrame:
    """Process the Wealth and Assets Survey household wealth file.

    Args:
            was_file (str): Path to the file: was_round_6_hhold_eul_mar_20.tab.

    Returns:
            pd.DataFrame: The processed dataframe.
    """
    RENAMES = {
        "R6xshhwgt": "weight",
        # Components for estimating land holdings.
        "DVLUKValR6_sum": "uk_land",
        "DVPropertyR6": "property_values",
        "DVFESHARESR6_aggr": "emp_shares_options",
        "DVFShUKVR6_aggr": "uk_shares",
        "DVIISAVR6_aggr": "investment_isas",
        "DVFCollVR6_aggr": "unit_investment_trusts",
        "TotpenR6_aggr": "pensions",
        "DvvalDBTR6_aggr": "db_pensions",
        # Predictors for fusing to FRS.
        "dvtotgirR6": "gross_income",
        "NumAdultW6": "num_adults",
        "NumCh18W6": "num_children",
        # Household Gross Annual income from occupational or private pensions
        "DVGIPPENR6_AGGR": "pension_income",
        "DVGISER6_AGGR": "self_employment_income",
        # Household Gross annual income from investments
        "DVGIINVR6_aggr": "investment_income",
        # Household Total Annual Gross employee income
        "DVGIEMPR6_AGGR": "employment_income",
        "HBedrmW6": "num_bedrooms",
        "GORR6": "region",
        "DVPriRntW6": "is_renter",  # {1, 2} TODO: Get codebook values.
        "CTAmtW6": "council_tax",
        # Other columns for reference.
        "DVLOSValR6_sum": "non_uk_land",
        "HFINWNTR6_Sum": "net_financial_wealth",
        "DVLUKDebtR6_sum": "uk_land_debt",
        "HFINWR6_Sum": "gross_financial_wealth",
        "TotWlthR6": "wealth",
    }

    was = (
        pd.read_csv(
            was_file, usecols=RENAMES.keys(), delimiter="\t", low_memory=False
        )
        .replace(" ", 0)
        .astype(float)
        .rename(columns=RENAMES)
    )

    was["is_renting"] = was["is_renter"] == 1

    # Land value held by households and non-profit institutions serving
    # households: 3.9tn as of 2019 (ONS).
    HH_NP_LAND_VALUE = 3_912_632e6
    # Land value held by financial and non-financial corporations.
    CORP_LAND_VALUE = 1_600_038e6
    # Land value held by government (not used).
    GOV_LAND_VALUE = 196_730e6

    was["non_db_pensions"] = was.pensions - was.db_pensions
    was["corp_wealth"] = was[
        [
            "non_db_pensions",
            "emp_shares_options",
            "uk_shares",
            "investment_isas",
            "unit_investment_trusts",
        ]
    ].sum(axis=1)

    totals = mdf.weighted_sum(
        was, ["uk_land", "property_values", "corp_wealth"], "weight"
    )

    land_prop_share = (
        HH_NP_LAND_VALUE - totals.uk_land
    ) / totals.property_values
    land_corp_share = CORP_LAND_VALUE / totals.corp_wealth

    was["est_land"] = (
        was.uk_land
        + was.property_values * land_prop_share
        + was.corp_wealth * land_corp_share
    )

    return was



from setuptools import setup, find_packages

setup(
    name="openfisca-uk-data",
    version="0.1.5",
    description=(
        "A Python package to manage OpenFisca-UK-compatible microdata"
    ),
    url="http://github.com/ubicenter/openfisca-uk-data",
    author="Nikhil Woodruff",
    author_email="nikhil.woodruff@outlook.com",
    packages=find_packages(exclude="microdata"),
    install_requires=[
        "pandas",
        "pathlib",
        "tqdm",
        "tables",
        "h5py",
        "google-cloud-storage",
        "jupyter-book>=0.11.1",
        "sphinxcontrib-bibtex>=1.0.0",
    ],
    entry_points={
        "console_scripts": ["openfisca-uk-data=openfisca_uk_data.cli:main"],
    },
)



from openfisca_uk_data.datasets.spi.raw_spi import RawSPI
from openfisca_core.model_api import *
from openfisca_uk_data.utils import *
import pandas as pd
from pandas import DataFrame
import h5py

max_ = np.maximum
where = np.where


@dataset
class SPI:
    name = "spi"
    model = UK

    def generate(year: int) -> None:
        """Generates the SPI-based input dataset for OpenFisca-UK.

        Args:
                year (int): The year to generate for (uses the raw SPI from this year)
        """

        main = RawSPI.load(year, "main").fillna(0)
        spi = h5py.File(SPI.file(year), mode="w")

        add_id_variables(spi, main)
        add_age(spi, main)
        add_incomes(spi, main)

        # Generate OpenFisca-UK variables and save
        spi.close()


def add_id_variables(spi: h5py.File, main: DataFrame):
    spi["person_id"] = main.index
    spi["person_benunit_id"] = main.index
    spi["person_household_id"] = main.index
    spi["benunit_id"] = main.index
    spi["household_id"] = main.index
    spi["role"] = np.array(["adult"] * len(main)).astype("S")


def add_age(spi: h5py.File, main: DataFrame):
    LOWER = np.array([16, 25, 35, 45, 55, 65, 75])
    UPPER = np.array([25, 35, 45, 55, 65, 75, 80])
    age_range = main.AGERANGE - 1
    spi["age"] = LOWER[age_range] + np.random.rand(len(main)) * (
        UPPER[age_range] - LOWER[age_range]
    )


def add_incomes(spi: h5py.File, main: DataFrame):
    RENAMES = dict(
        pension_income="PENSION",
        self_employment_income="PROFITS",
        property_income="INCPROP",
        savings_interest_income="INCBBS",
        dividend_income="DIVIDENDS",
        blind_persons_allowance="BPADUE",
        married_couples_allowance="MCAS",
        gift_aid="GIFTAID",
        capital_allowances="CAPALL",
        deficiency_relief="DEFICIEN",
        covenanted_payments="COVNTS",
        charitable_investment_gifts="GIFTINV",
        employment_expenses="EPB",
        other_deductions="MOTHDED",
        pension_contributions="PENSRLF",
        person_weight="FACT",
        benunit_weight="FACT",
        household_weight="FACT",
    )
    spi["pays_scottish_income_tax"] = main.SCOT_TXP == 1
    spi["employment_income"] = main[["PAY", "EPB", "TAXTERM"]].sum(axis=1)
    spi["social_security_income"] = main[
        ["SRP", "INCPBEN", "UBISJA", "OSSBEN"]
    ].sum(axis=1)
    spi["miscellaneous_income"] = main[
        ["OTHERINV", "OTHERINC", "MOTHINC"]
    ].sum(axis=1)
    for var, key in RENAMES.items():
        spi[var] = main[key]



from openfisca_uk_data.utils import *
import synthimpute as si
import numpy as np
import h5py


@dataset
class FRS_SPI_Adjusted:
    name = "frs_spi_adj"
    model = UK

    def generate(year):
        from openfisca_uk import Microsimulation
        from openfisca_uk_data.datasets.frs.frs import FRS
        from openfisca_uk_data.datasets.spi import SPI

        frs_sim = Microsimulation(dataset=FRS, year=year)
        spi_sim = Microsimulation(dataset=SPI, year=year)

        X_spi = np.array(
            [
                spi_sim.calc("employment_income", year).values,
                spi_sim.calc("pension_income", year).values,
                spi_sim.calc("age", year).values,
            ]
        ).T
        X_frs = np.array(
            [
                frs_sim.calc("employment_income", year).values,
                frs_sim.calc("pension_income", year).values,
                frs_sim.calc("age", year).values,
            ]
        ).T
        Y_spi = spi_sim.calc("dividend_income", year).values
        spi_weight = spi_sim.calc("person_weight", year).values
        frs_weight = frs_sim.calc("person_weight", year).values
        print(
            "Imputing dividend income for FRS respondents from SPI values...",
            end="",
        )
        imputed_dividend_income = si.rf_impute(
            x_train=X_spi,
            y_train=Y_spi,
            x_new=X_frs,
            sample_weight_train=spi_weight,
            new_weight=frs_weight,
            target=spi_sim.calc("dividend_income").sum(),
        )
        print(" completed.")
        frs_sim.simulation.set_input(
            "dividend_income", year, imputed_dividend_income
        )
        with h5py.File(FRS_SPI_Adjusted.file(year), "w") as f:
            for variable in MAIN_INPUT_VARIABLES:
                f[f"{variable}/{year}"] = frs_sim.calc(variable, year).values



from argparse import ArgumentParser
from openfisca_uk_data import *


def main():
    datasets = {
        ds.name: ds
        for ds in (
            RawFRS,
            UKMODInput,
            UKMODOutput,
            FRS,
            SynthFRS,
            RawSPI,
            SPI,
            FRS_SPI_Adjusted,
            FRS_WAS_Imputation,
        )
    }
    parser = ArgumentParser(
        description="A utility for storing OpenFisca-UK-compatible microdata."
    )
    parser.add_argument("dataset", help="The dataset to select")
    parser.add_argument("action", help="The action to take")
    parser.add_argument(
        "args", nargs="*", help="The arguments to pass to the function"
    )
    args = parser.parse_args()
    try:
        return getattr(datasets[args.dataset], args.action)(*args.args)
    except Exception as e:
        print(f"Encountered an error: {e.with_traceback()}")


if __name__ == "__main__":
    main()
